﻿Type=Service
Version=3.82
B4A=true
@EndOfDesignText@
#Region Module Attributes
	#StartAtBoot: False
#End Region

'Service module
Sub Process_Globals
	Dim WifiServer As ServerSocket
	Dim BluetoothServer As Serial
	Dim Client As Socket
	Dim Streams As AsyncStreams
	Dim Notification1 As Notification
	Dim PE As PhoneEvents
	Dim Phone As Phone
	Dim PhoneId As PhoneId
	Dim LogCat As LogCat
	Dim BC As ByteConverter
	Dim ConnectedStatus As Boolean
	Dim Out As OutputStream
	Dim LastToastMessageShown As Long
	Dim	PING = 1, APK_COPY_START = 2, APK_PACKET = 3 _
		,APK_COPY_CLOSE = 4, SEND_DEVICE_NAME = 5, START_LOGCAT = 6 _
		,STOP_LOGCAT = 7, LOGCAT_DATA = 8, LAUNCH_PROGRAM = 9, KILL_PROGRAM = 10 _
		,ASK_FOR_DESIGNER = 11, BRIDGE_LOG_PORT = 12 As Byte
	Dim DisconnectTimer As Timer
	Dim DisconnectTicks As Int
	Dim wifi As Boolean
	Private apkName As Int = 1
	Private noDesignerInstalled As Boolean
	Private lastCheckForDesigner As Long
	Private UdpServer As UDPSocket
End Sub
Sub Service_Create
	Try
		If WifiServer.IsInitialized = False Then
			WifiServer.Initialize(6789, "Server")
		End If
	Catch
		Log(LastException.Message)
		'switch to alternate port
		WifiServer.Initialize(6789 + 117, "Server")
	End Try
	If Phone.SdkVersion >= 16 Then 'workaround for new security feature of Jelly Bean
		If UdpServer.IsInitialized = False Then
			UdpServer.Initialize("Udp", 0, 8192)
		End If
	End If
	BluetoothServer.Initialize("BluetoothServer")
	DisconnectTimer.Initialize("DisconnectTimer", 1000)
	Notification1.Initialize
	Notification1.Icon = "icon"
	Notification1.Light = False
	Notification1.Sound = False
	Notification1.Vibrate = False
	Notification1.SetInfo("B4A-Bridge", "Disconnected", Main)
	Service.StartForeground(1, Notification1)
	PE.Initialize("PE")
End Sub

Sub Service_Start
	Try
		UpdateStatus (ConnectedStatus, True)
	Catch
		ShowToastMessage("Error starting service: " & LastException.Message)
	End Try
End Sub

Sub UpdateStatus (Connected As Boolean, AllowFurtherConnections As Boolean)
	ConnectedStatus = Connected
	If Connected Then
		DisconnectTimer.Enabled = True
		DisconnectTicks = 0
		Notification1.SetInfo("B4A-Bridge", "Connected", Main)
		CheckIfDesignerIsInstalled
	Else
		If Streams.IsInitialized Then Streams.Close
		DisconnectTimer.Enabled = False
		Notification1.SetInfo("B4A-Bridge", "Disconnected", Main)
	End If
	If AllowFurtherConnections Then
		'Service should already be initialized here.
		Notification1.Notify(1)
		If wifi Then
			WifiServer.Listen
		Else
			BluetoothServer.Listen2("b4a-bridge", "fa8fc0d0-afac-11de-8a39-08f0200c9a66")
		End If
	Else
		BluetoothServer.Disconnect
	End If
	CallSubDelayed(Main, "UpdateStatus")
End Sub
Sub DisconnectTimer_Tick
	DisconnectTicks = DisconnectTicks + 1
	If DisconnectTicks = 10 Then
		UpdateStatus(False, True)
	End If
End Sub
Sub Service_Destroy
	Service.StopForeground(1)
	UpdateStatus (False, False)
	LogCat.LogCatStop
End Sub
Sub BluetoothServer_Connected (Success As Boolean)
	If Streams.IsInitialized Then Streams.Close
	Dim Streams As AsyncStreams 'create a new streams object.
	If Success Then
		Log("Connected to B4A-Bridge (Bluetooth)")
		Streams.InitializePrefix(BluetoothServer.InputStream, False, BluetoothServer.OutputStream, _
			"Streams")
		AfterConnect
	Else
		ShowToastMessage(LastException.Message)
		UpdateStatus(False, True)
	End If
End Sub

Sub Server_NewConnection (Successful As Boolean, NewSocket As Socket)
	If wifi = False Then 
		NewSocket.Close
		Return
	End If
	If Streams.IsInitialized Then Streams.Close
	Dim Streams As AsyncStreams 'create a new streams object.
	If Successful Then
		Log("Connected to B4A-Bridge (Wifi)")
		Client = NewSocket
		Streams.InitializePrefix(Client.InputStream, False, Client.OutputStream, _
			"Streams")
		AfterConnect
	Else
		ShowToastMessage(LastException.Message)
		UpdateStatus(False, True)
	End If
End Sub
Sub AfterConnect
	Dim pingB(1) As Byte
	pingB(0) = PING
	Streams.Write(pingB)
	Dim name As String
	name = Phone.Manufacturer & " " & Phone.Model & "-" & PhoneId.GetDeviceId
	Dim nameBytes() As Byte
	nameBytes = name.GetBytes("UTF8")
	Streams.Write(AddCommandToBytes(SEND_DEVICE_NAME, nameBytes, nameBytes.Length))
	UpdateStatus(True, True)
End Sub
Sub Streams_Error
	If Sender <> Streams Then Return
	ShowToastMessage(LastException.Message)
	UpdateStatus(False, True)
End Sub
Sub Streams_Terminated
	If Sender <> Streams Then Return
	UpdateStatus(False, True)
	Log("Streams_terminated")
End Sub


Sub AddCommandToBytes(Command As Byte, Buffer() As Byte, Length As Int) As Byte()
	Dim B(Length + 1) As Byte
	B(0) = Command
	BC.ArrayCopy(Buffer, 0, B, 1, Length)
	Return B
End Sub

Sub Streams_NewData (Buffer() As Byte)
	If Streams.IsInitialized = False Then Return
	DisconnectTicks = 0 'reset the disconnect timer
	Dim command As Byte
	command = Buffer(0)
	Select command
		Case PING
			Dim B(1) As Byte
			B(0) = PING
			Streams.Write(B)
			If noDesignerInstalled AND DateTime.Now > lastCheckForDesigner + 20 * DateTime.TicksPerSecond Then CheckIfDesignerIsInstalled
		Case APK_COPY_START
			HandleFileStart
		Case APK_PACKET
			HandleFilePacket(Buffer)
		Case APK_COPY_CLOSE
			HandleFileClose
		Case START_LOGCAT
			StartLogcat(Buffer)
		Case STOP_LOGCAT
			StopLogcat
		Case LAUNCH_PROGRAM
			LaunchProgram(Buffer)
		Case KILL_PROGRAM
			KillProgram(Buffer)
	End Select
End Sub
Sub CheckIfDesignerIsInstalled
	Dim pm As PackageManager
	Dim packages As List
	packages = pm.GetInstalledPackages
	Dim buffer() As Byte
	If packages.IndexOf("anywheresoftware.b4a.designer") = -1 Then
		buffer = "0".GetBytes("UTF8")
		noDesignerInstalled = True
		lastCheckForDesigner = DateTime.Now
	Else
		noDesignerInstalled = False
		Dim s As String
		s = pm.GetVersionCode("anywheresoftware.b4a.designer")
		buffer = s.GetBytes("UTF8")
	End If
	
	'we are sending the installed designer version so the IDE will decide if a new version should be installed.
	Streams.Write(AddCommandToBytes(ASK_FOR_DESIGNER, buffer, buffer.Length))
End Sub
Sub LaunchProgram (Buffer() As Byte)
	Dim s As String
	s = BytesToString(Buffer, 1, Buffer.Length - 1, "UTF8")
	Dim args() As String
	args = Regex.Split(",", s)
	Dim In1 As Intent
	For i = 1 To args.Length - 1 Step 2
		Select args(i)
			Case "-a"
				In1.Initialize(args(i + 1), "")
			Case "-n"
				In1.SetComponent(args(i + 1))
			Case "-f"
				In1.Flags = Bit.ParseInt(args(i + 1).SubString(2), 16)
			Case "-c"
				In1.SetComponent(args(i + 1))
			Case "-e"
				In1.PutExtra(args(i + 1), args(i + 2))
				Exit
		End Select
	Next
	Log(In1)
	Log(In1.ExtrasToString)
	WriteBluetoothFile
	Try
		StartActivity(In1)
	Catch
		ShowToastMessage("Error starting intent: " & In1) 'ignore
	End Try
End Sub
Sub KillProgram (Buffer() As Byte)
	Dim package As String
	package = BytesToString(Buffer, 1, Buffer.Length - 1, "UTF8")
	Dim sb As StringBuilder
	sb.Initialize
	Phone.Shell("ps", Null, sb, Null)
	Dim m As Matcher
	m = Regex.Matcher2("^[^ ]*\s+(\d+) .*" & package, Regex.MULTILINE, sb.ToString)
	If m.Find Then
		Log("Package found: " & package)
		Phone.Shell("kill", Array As String(m.Group(1)), Null, Null)
	End If
End Sub
Sub StartLogcat (Buffer() As Byte)
	Dim args As String
	args = BytesToString(Buffer, 1, Buffer.Length - 1, "UTF8")
	LogCat.LogCatStart(Regex.Split(",", args), "LogCat")
End Sub

Sub LogCat_LogCatData (Buffer() As Byte, Length As Int)
	If ConnectedStatus = True Then
		Streams.Write(AddCommandToBytes(LOGCAT_DATA, Buffer, Length))
	End If
End Sub
Sub StopLogcat
	LogCat.LogCatStop
End Sub
Sub HandleFileStart
	Log("Installing file.")
	If Out.IsInitialized Then Out.Close
	Try
		Out = File.OpenOutput(Main.folder, "temp" & apkName & ".apk", False)
	Catch
		Log(LastException.Message) 'this can happen if the file is locked.
		apkName = apkName + 1
		Out = File.OpenOutput(Main.folder, "temp" & apkName & ".apk", False)
	End Try
End Sub
Sub HandleFilePacket(Buffer() As Byte)
	Try
		Out.WriteBytes(Buffer, 1, Buffer.Length - 1)
	Catch
		Log(LastException)
	End Try
End Sub

Sub HandleFileClose
	Out.Close
	WriteBluetoothFile
	'ask the system to install the apk
	Dim i As Intent
	i.Initialize(i.ACTION_VIEW, "file://" & File.Combine(Main.folder, "temp" & apkName & ".apk"))
	i.SetType("application/vnd.android.package-archive")
	StartActivity(i)
	If UdpServer.IsInitialized Then
		Dim port As String
		port = UdpServer.port
		Dim b() As Byte
		b = port.GetBytes("UTF8")
		Streams.Write(AddCommandToBytes(BRIDGE_LOG_PORT, b, b.Length))
	End If
End Sub
Sub WriteBluetoothFile
	If wifi = False Then
		File.WriteString(File.DirRootExternal, "b4a_bluetooth.txt", "0")
	Else
		Try
			File.Delete(File.DirRootExternal, "b4a_bluetooth.txt")
		Catch
			Log("Failed to delete file.")
		End Try
	End If
End Sub
Sub PE_PackageAdded (Package As String, Intent As Intent)
	Log("PackageAdded: " & Package)
End Sub
Sub PE_ConnectivityChanged (NetworkType As String, State As String, Intent As Intent)
	If NetworkType = "WIFI" Then CallSub(Main, "UpdateIp")
End Sub
Sub ShowToastMessage(s As String)
	Log(s)
	If LastToastMessageShown + 5 * DateTime.TicksPerSecond > DateTime.Now Then
		Return
	Else
		LastToastMessageShown = DateTime.Now
		ToastMessageShow(s, True)
	End If
End Sub
Sub UDP_PacketArrived (Packet As UDPPacket)
	Log(BytesToString(Packet.Data, Packet.Offset, Packet.Length, "UTF8"))
End Sub
